/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 *
 * Copyright (©) 2017,2018 Marc Alexander Lehmann / the Deliantra team
 * Copyright (©) 2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2002 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992 Frank Tore Johansen
 *
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#include <stdio.h>
#include <global.h>

sint64 levels[MAXNUMLEVELS];

static const float exp_att_mult[NROFATTACKS + 2] = {
  0.0,                          /* AT_PHYSICAL  */
  0.0,                          /* AT_MAGIC     */
  0.0,                          /* AT_FIRE      */
  0.0,                          /* AT_ELECTRICITY */
  0.0,                          /* AT_COLD      */
  0.0,                          /* AT_WATER     *//*AT_CONFUSION! */
  0.4,                          /* AT_ACID      */
  1.5,                          /* AT_DRAIN     */
  0.0,                          /* AT_WEAPONMAGIC */
  0.1,                          /* AT_GHOSTHIT  */
  0.3,                          /* AT_POISON    */
  0.2,                          /* AT_DISEASE   */
  0.3,                          /* AT_PARALYZE  */
  0.0,                          /* AT_TURN_UNDEAD */
  0.0,                          /* AT_FEAR      */
  0.0,                          /* AT_CANCELLATION */
  0.0,                          /* AT_DEPLETE */
  0.0,                          /* AT_DEATH */
  0.0,                          /* AT_CHAOS */
  0.0                           /* AT_COUNTERSPELL */
};

static const float exp_prot_mult[NROFATTACKS + 2] = {
  0.4,                          /* AT_PHYSICAL  */
  0.5,                          /* AT_MAGIC     */
  0.1,                          /* AT_FIRE      */
  0.1,                          /* AT_ELECTRICITY */
  0.1,                          /* AT_COLD      */
  0.1,                          /* AT_WATER     */
  0.1,                          /* AT_ACID      */
  0.1,                          /* AT_DRAIN     */
  0.1,                          /* AT_WEAPONMAGIC */
  0.1,                          /* AT_GHOSTHIT  */
  0.1,                          /* AT_POISON    */
  0.1,                          /* AT_DISEASE   */
  0.1,                          /* AT_PARALYZE  */
  0.1,                          /* AT_TURN_UNDEAD */
  0.1,                          /* AT_FEAR      */
  0.0,                          /* AT_CANCELLATION */
  0.0,                          /* AT_DEPLETE */
  0.0,                          /* AT_DEATH */
  0.0,                          /* AT_CHAOS */
  0.0                           /* AT_COUNTERSPELL */
};

/*
 * Returns true if the monster specified has any innate abilities.
 */
static int
has_ability (const object *ob)
{
  for (object *tmp = ob->inv; tmp; tmp = tmp->below)
    if (tmp->type == SPELL || tmp->type == SPELLBOOK)
      return true;

  return false;
}

/*
 * new_exp() is an alternative way to calculate experience based
 * on the ability of a monster.
 * It's far from perfect, and doesn't consider everything which
 * can be considered, thus it's only used in debugging.
 * this is only used with one of the dumpflags,
 * and not anyplace in the code.
 */
int
new_exp (const object *ob)
{
  double att_mult, prot_mult, spec_mult;
  double exp;
  int i;
  long mask = 1;

  att_mult = prot_mult = spec_mult = 1.0;
  for (i = 0; i < NROFATTACKS; i++)
    {
      mask = 1 << i;
      att_mult += (exp_att_mult[i] * ((ob->attacktype & mask) != 0));
      /* We multiply & then divide to prevent roundoffs on the floats.
       * the doubling is to take into account the table and resistances
       * are lower than they once were.
       */
      prot_mult += (exp_prot_mult[i] * 200 * ob->resist[i]) / 100.0;
    }

  spec_mult += (0.3 * (ob->flag [FLAG_SEE_INVISIBLE] != 0)) +
    (0.5 * (ob->flag [FLAG_SPLITTING] != 0)) +
    (0.3 * (ob->flag [FLAG_HITBACK] != 0)) +
    (0.1 * (ob->flag [FLAG_REFL_MISSILE] != 0)) +
    (0.3 * (ob->flag [FLAG_REFL_SPELL] != 0)) +
    (1.0 * (ob->flag [FLAG_NO_MAGIC] != 0)) +
    (0.1 * (ob->flag [FLAG_PICK_UP] != 0)) +
    (0.1 * (ob->flag [FLAG_USE_SCROLL] != 0)) +
    (0.2 * (ob->flag [FLAG_USE_RANGE] != 0)) + (0.1 * (ob->flag [FLAG_USE_BOW] != 0));

  exp = (ob->stats.maxhp < 5) ? 5 : ob->stats.maxhp;
  exp *= (ob->flag [FLAG_CAST_SPELL] && has_ability (ob)) ? (40 + (ob->stats.maxsp > 80 ? 80 : ob->stats.maxsp)) / 40 : 1;
  exp *= (80.0 / (70.0 + ob->stats.wc)) * (80.0 / (70.0 + ob->stats.ac)) * (50.0 + ob->stats.dam) / 50.0;
  exp *= att_mult * prot_mult * spec_mult;
  exp *= 2.0 / (2.0 - min (ob->speed, 0.95));
  exp *= (20.0 + ob->stats.Con) / 20.0;
  if (ob->flag [FLAG_STAND_STILL])
    exp /= 2;

  return exp;
}

//TODO: binary search...
int
exp_to_level (sint64 exp)
{
  for (int i = 1; i <= settings.max_level; i++)
    if (levels [i] > exp)
      return i - 1;

  return settings.max_level;
}

sint64
level_to_min_exp (int level)
{
  return levels [clamp (level, 0, settings.max_level)];
}

/* This loads the experience table from the exp_table
 * file.
 */
void
_reload_exp_table ()
{
  int lastlevel = 0;
  sint64 lastexp = -1;

  object_thawer thawer (settings.datadir, "exp_table");

  if (!thawer)
    {
      LOG (llevError, "unable to load experience table file");
      return;
    }

  if (thawer.kw != KW_max_level)
    {
      thawer.parse_error ("experience table file");
      return;
    }

  thawer.get (settings.max_level);

  sint64 newlevels [MAXNUMLEVELS];
  newlevels [0] = 0;

  while (thawer.next_line ())
    {
      sint64 tmpexp;
      thawer.get (tmpexp);

      /* Do some sanity checking - if value is bogus, just exit because
       * the table otherwise is probably in an inconsistent state
       */
      if (tmpexp <= lastexp)
        {
          LOG (llevError, "Experience for level %d is lower than previous level (%" PRId64 " <= %" PRId64 ")\n", lastlevel + 1, tmpexp, lastexp);
          return;
        }

      lastlevel++;

      if (lastlevel > settings.max_level || lastlevel >= MAXNUMLEVELS - 1)
        {
          LOG (llevError, "Too many levels specified in table (%d > %d (settings.max_level) or >= %d (MAXNUMLEVELS - 1))\n", lastlevel, settings.max_level, MAXNUMLEVELS - 1);
          exit (1);
        }

      newlevels [lastlevel] = tmpexp;
      lastexp = tmpexp;
    }

  if (lastlevel != settings.max_level && lastlevel != 0)
    {
      LOG (llevError, "Warning: exp_table does not have %d entries (%d)\n", settings.max_level, lastlevel);
      return;
    }

  // copy the last exp value one level higher, so we can take the difference to the "next" level
  // eve4n for players at max. level
  newlevels [lastlevel + 1] = newlevels [lastlevel];

  memcpy (levels, newlevels, sizeof (levels));
}

